from xevo.crossevo import crossevo

from xevo.optimizers.morthoevo import morthoevo
from xevo.optimizers.orthoevo import orthoevo
from xevo.optimizers.trajmuteplus import trajmuteplus
from xevo.optimizers.trajmute import trajmute
from xevo.optimizers.trivmute import trivmute
from xevo.optimizers.trivmutetraj import trivmutetraj



import matplotlib.pyplot as plt
import numpy as np
import time


from xevo.erun import erun

class erunplus(erun):
    
    def __init__(s,vo,obj,population=10,show=True,delay=None,custom_cancel=None):
        if custom_cancel is None:custom_cancel=lambda **kwargs:False
        s.custom_cancel=custom_cancel
        erun.__init__(s,vo,obj,population=population,show=show,delay=delay)
    
    def run(s,maxsteps=1000000,goalstrength=1000000.0):
        for i in range(maxsteps):
            mx,mm,ss=s.vo.logeneration(s.show)
            s.maxs.append(mx)
            s.means.append(mm)
            s.stds.append(ss)
            if s.custom_cancel(mx=mx,mm=mm,ss=ss,s=s):return
            if s.shallmaximize:
                if mx>=goalstrength:return
            else:
                if mx<=goalstrength:return
            if not s.delay is None:
                time.sleep(s.delay)
    
            











from xevo.evo import evo

import numpy as np
import time
import random

class trivmute(evo):
  """take x, if 50%->mutate (if better->set x, else ignore) else randomize"""
  
  def initial(s):
    s.q=None
    s.i=0
  def populate(s,person,n=10)->None:
    s.q=person.copy()#.randomize()
  
  def winner(s)->int:
    return 0
  def topn(s,n=3)->"[int]":
    return [0]
  def average(s)->"float,float":
    return s.q.strength(),0
  
  def __getitem__(s,key):
    return s.q
  def __set_item__(s,key,obj):
    s.q=obj
  
  def copy(s):
    ret=s._copy()
    if not s.q is None:ret.q=s.q.copy()
    return ret
  
  def __init__(s):
    s.initial()



  def generation(s)->None:
    os=s.q.strength()
    if random.random()<0.5:
      n=s.q.randomize()
    else:
      n=s.q.mutate()
    ns=n.strength()
    if ns<=os:
      s.q=n
      if hasattr(s.q,"getsave"):s.q.getsave()("data/bestobj.txt")
    
    


  def _copy(s):
    return trivmute()






from xevo.evo import evo

import numpy as np
import time
        

class orthoevo(evo):
    """similar to lineevo and crossevo but uses oobj to define neighbourhood as similarity. Can only work with 1d similarity and ignores everything else. Works, but the drawback is that this is a lot slower (ignoring the sort, this requires still 2x-4x epochs)"""
    
    def __init__(s,wsmerge=0.3,dex=0):
        s.initial()
        s.wsmerge=wsmerge
        s.dex=dex



    def generation(s)->None:
        s.q.sort(key=lambda x:x.ortho()[s.dex])
        i1=np.random.randint(len(s.q)-1)
        i2=i1+1
        i1won=s.q[i1].figth(s.q[i2])
        iwon=i1
        ilost=i2
        if not i1won:
            iwon=i2
            ilost=i1
        if np.random.random()<s.wsmerge:
            s.q[ilost]+=s.q[iwon]
        else:
            s.q[ilost]=s.q[iwon].mutate()
        
        


    def _copy(s):
        return orthoevo(s.wsmerge,s.dex)






from xevo.eobj import eobj

import numpy as np

class pion(eobj):
  """trivial test eobj trying to find the exact value of pi"""
  def __init__(s,q=3):
    s.initial()
    s.q=q
  def __str__(s):
    return str(s.q)
  def __add__(a,b):
    return pion((a.q+b.q)/2)
  def shallmaximize(s):return False

  
  def randomize(s):
    return pion(np.random.random()*2+2)
  def mutate(s):
    sn=s.q
    alpha=2*np.random.random()-1
    power=-10*np.random.random()
    sn+=alpha*np.exp(power)
    return pion(sn)
  def calcstrength(s):
    return np.abs(s.q-np.pi)
  def _copy(s):
    return pion(s.q)



from xevo.eobj import eobj

import numpy as np

class scounter:
  def __init__(s):
    s.i=0
  def pp(s):
    s.i+=1
  def __repr__(s):
    return str(s)
  def __str__(s):
    return str(s.i)
  def reset(s):
    s.i=0

counter=scounter()


class bitflip(eobj):
  """trivial test eobj trying to maximize the sum of a list of 100 booleans. Also counts each call to calcstrength"""
  def __init__(s,q=None):
    s.initial()
    s.q=q
    if q is None:s.q=s.randomize().q
  def __str__(s):
    return "".join([str(qq) for qq in s.q])
  def __add__(a,b):
    ret=[]
    for aa,bb in zip(a.q,b.q):
      if np.random.randint(2)==0:
        ret.append(aa)
      else:
        ret.append(bb)
    return bitflip(ret)
  
  def randomize(s):
    ret=[]
    for i in range(100):
      ret.append(np.random.randint(2))
    return bitflip(ret)
  def mutate(s):
    rel=[p for p in s.q]
    i=np.random.randint(len(rel))
    rel[i]=1-rel[i]
    return bitflip(rel)
  def calcstrength(s):
    counter.pp()
    return float(np.sum(s.q))
  def _copy(s):
    return bitflip([p for p in s.q])



from xevo.crossevo import crossevo
from xevo.erun import erun

from xevo.erunplus import erunplus


def quickmut(obj,population=20,goal=0.0,maxsteps=1000,show=True,opt=None):

    if opt is None:opt=crossevo()

    c=erun(opt,obj,population=population,show=show)
    c.run(goalstrength=goal,maxsteps=maxsteps)
    return c.getwinner()


def semiquickmut(obj,population=20,goal=0.0,maxsteps=1000,show=True,opt=None):

    if opt is None:opt=crossevo()

    c=erun(opt,obj,population=population,show=show)
    c.run(goalstrength=goal,maxsteps=maxsteps)
    return c.vo.q



def meanquickmut(obj,population=20,goal_max=0.0,goal_mean=0.0,maxsteps=1000,show=True,opt=None):

    def cancel(s,mm,**kwargs):
        if s.shallmaximize:
            return mm>goal_mean
        else:
            return mm<goal_mean


    if opt is None:opt=crossevo()

    c=erunplus(opt,obj,population=population,show=show,custom_cancel=cancel)
    c.run(goalstrength=goal_max,maxsteps=maxsteps)
    return c.vo.q







from xevo.evo import evo

import numpy as np
import time
import random

class trajmuteplus(evo):
    def __init__(s):
        s.initial()



    def generation(s)->None:
        s.q=[zw.mutate() for zw in s.q]
        i1=np.random.randint(len(s.q))
        i2=i1
        #print(i1,i2,len(s.q))
        while i2==i1 and (not len(s.q)<2):
            i2=np.random.randint(len(s.q))
        s1=s.q[i1].strength()
        s2=s.q[i2].strength()
        if not s.q[i1].shallmaximize():s1*=-1
        if not s.q[i2].shallmaximize():s2*=-1
        if s1<s2:
            s.q[i1]=s.q[i2]
        else:
            s.q[i2]=s.q[i1]


    

    def _copy(s):
        return trajmuteplus()






from xevo.eobj import eobj

from abc import abstractmethod,ABCMeta


class oobj(eobj):
    """equivalent zu eobj, but contains also functions to define similarity"""


    def ortho(s) -> "returns [dimensions], number defined in orthodim(s)":
        if (not hasattr(s,"oo")) or s.oo is None:
            s.oo=s.calcortho()
            if type(s.oo)!=list:
                s.oo=[s.oo]
        return s.oo

    @abstractmethod
    def calcortho(s):
        """calculate index values that define similarity"""
        pass

    def orthodim(s)->"dimensionality of the output":return 1

    def copy(s)->"subclass of oobj":
        ret=eobj.copy(s)
        #if hasattr(s,"oo"):
        #    ret.oo=s.oo
        return ret







from xevo.oobj import oobj

import random

class mob(oobj):
    """multiple oobj are also an oobj"""

    def __init__(s,gen_new=None,d=3,mutws=1.0,shallrnd=True):
        s.initial()
        s.d=d
        s.mutws=mutws

        if gen_new is None:return#allow initialisations without generation
        s.gen_new=gen_new

        s.q=[gen_new() for i in range(d)]

        s.odim=s.q[0].orthodim()*d

        if shallrnd:s.q=s.randomize().q


    def __str__(s):
        return "|".join([str(zw) for zw in s.q])

    def __add__(a,b):
        ret=a.copy()
        for i in range(ret.d):
            ret.q[i]+=b.q[i]
        return ret


    def shallmaximize(s):
        return s.q[0].shallmaximize()

    def randomize(s):
        ret=s.copy()
        for i in range(len(ret.q)):
            ret.q[i]=ret.q[i].randomize()
        return ret
    def mutate(s):
        ret=s.copy()
        for i in range(len(ret.q)):
            if random.random()>s.mutws:continue
            ret.q[i]=ret.q[i].mutate()
        return ret
    def calcstrength(s):
        p=1.0
        for zw in s.q:
            p*=zw.strength()
        return p**(1/s.d)
    def _copy(s):
        ret=mob()
        ret.gen_new=s.gen_new
        ret.d=s.d
        ret.q=[zw.copy() for zw in s.q]
        ret.odim=s.odim
        ret.mutws=s.mutws

        return ret

    def orthodim(s):
        return s.odim
    def calcortho(s):
        ret=[]
        #print("calc ortho")
        for zw in s.q:
            #print(zw)
            for zx in zw.ortho():
                ret.append(zx)
        return ret
   















from abc import ABCMeta,abstractmethod

import numpy as np


class eobj(object):
  """
  any eobj class contains at least
  def __add__(a,b):
  def randomize(s):
  def mutate(s):
  def calcstrength(s):
  def _copy(s):
  
  also __init__(s) should not contain any nonoptional parameters and call s.initial()
  """
  __metaclass__=ABCMeta
  
  def initial(s):
    s.stre=None
  
  def figth(a,b):
    """does a win against b"""
    sa=a.strength()
    sb=b.strength()
    
    if not a.shallmaximize():sa*=-1
    if not b.shallmaximize():sb*=-1
    if not sa==sb:return sa>sb
    return np.random.randint(2)==0
    
  
  def shallmaximize(s):
    """can be flipped to allow for minimization tasks"""
    return True
  
  @abstractmethod
  def __add__(a,b) -> "subclass of eobj":
    """define how to combine two objects"""
    pass
  
  @abstractmethod
  def randomize(s):
    """initialises a new object with initial random? parameters"""
    pass

  @abstractmethod
  def mutate(s) ->"subclass of eobj":
    """define how to mutate an object"""
    pass
  
  @abstractmethod
  def calcstrength(s) -> float:
    """calculate how strong the current object is"""
    return 0.0
  
  def strength(s) -> float:
    """define how strong the current object is"""
    if (not hasattr(s, 'stre')) or s.stre is None:
      s.stre=s.calcstrength()
    return s.stre
  

  @abstractmethod
  def _copy(s) -> "subclass of eobj":
    """copy the current object"""
  
  def copy(s) -> "subclass of eobj":
    """copy the current object"""
    ret=s._copy()
    return ret
  
  def __repr__(s):
    return str(s)


from xevo.eobj import eobj
from xevo.evo import evo
from xevo.erun import erun
from xevo.crossevo import crossevo

from xevo.quickmut import quickmut,semiquickmut,meanquickmut
from xevo.eobjplus import eobjplus
from xevo.oobj import oobj
from xevo.phobia import phobia
from xevo.erunplus import erunplus
from xevo.mob import mob


import xevo.objects
import xevo.optimizers




from xevo.eobj import eobj

import random

class phobia(eobj):
  pmute=0.3
  """implementing an phobic evolutionary object, requires an eobjplus"""
  def __init__(s,q=None,phob=None):
    s.initial()
    s.q=q
    if phob is None:phob=[]
    s.p=phob
  def __str__(s):
    return str(s.q)
  def __add__(a,b):
    return phobia(a.q+b.q,random.choice([a.p,b.p]))
  def shallmaximize(s):s.q.shallmaximize()

  
  def randomize(s):
    return phobia(s.q.randomize(),[])
  
  def alterpm(s):
    le=len(s.q.func)
    i=random.random()*le/1.0001
    if i in s.p:
      s.p.remove(i)
    else:
      s.p.append(i)
    while len(s.p)>=le:
      s.p=s.p[1:]
  def mutate(s):
    if random.random()<s.pmute:
      s.alterpm()
    # print("gonna mutate",s.q.func)
    return phobia(s.q._mutbyfunc([q for i,q in enumerate(s.q.func) if not i in s.p]),[i for i in s.p])
    
    
  def calcstrength(s):
    return s.q.calcstrength()
  def _copy(s):
    return phobia(s.q.copy(),[i for i in s.p])





from xevo.eobj import eobj
from xevo.oobj import oobj

import numpy as np

class oion(oobj):
    """modification of pion, to result in multiplicates of pi (for orthogonal optimisation)"""
    def __init__(s,q=3):
        s.initial()
        s.q=q
    def __str__(s):
        return str(s.q)
    def __add__(a,b):
        return oion((a.q+b.q)/2)
    def shallmaximize(s):return False

    
    def randomize(s):
        return oion(np.random.random()*100)
    def mutate(s):
        sn=s.q
        alpha=2*np.random.random()-1
        power=-10*np.random.random()+1#+1 compared to pion, since bigger range
        sn+=alpha*np.exp(power)
        return oion(sn)
    def calcstrength(s):
        return np.abs(np.sin(s.q))
    def _copy(s):
        return oion(s.q)
    def calcortho(s):
        return s.q



from xevo.eobj import eobj

from xevo.oobj import oobj


import numpy as np



class dualflip(oobj):
    """trivial test eobj trying to maximize the sum of a list of 100 booleans."""
    def __init__(s,q=None):
        s.initial()
        s.q=q
        if q is None:s.q=s.randomize().q
    def __str__(s):
        return "".join([str(qq) for qq in s.q])
    def __add__(a,b):
        ret=[]
        for aa,bb in zip(a.q,b.q):
            if np.random.randint(2)==0:
                ret.append(aa)
            else:
                ret.append(bb)
        return dualflip(ret)
    
    def randomize(s):
        ret=[]
        for i in range(100):
            ret.append(np.random.randint(2))
        return dualflip(ret)
    def mutate(s):
        rel=[p for p in s.q]
        i=np.random.randint(len(rel))
        rel[i]=1-rel[i]
        return dualflip(rel)
    def calcstrength(s):
        return abs(2*(float(np.sum(s.q))-50))/100#lets define it on the more usual range
    def _copy(s):
        return dualflip([p for p in s.q])

    def calcortho(s):
        return float(np.mean(s.q))

    def orthodim(s):return 1






    




from xevo.evo import evo

import numpy as np
import time
import random

class trajmute(evo):
    def __init__(s):
        s.initial()



    def generation(s)->None:
        for i in range(len(s.q)):
            s.q[i]=s.q[i].mutate()
        stre=[zw.strength()* (1 if zw.shallmaximize() else -1)  for zw in s.q]
        wori=np.argmin(stre)
        s.q[wori]=np.random.choice(s.q,p=[float(wori!=i)/(len(s.q)-1) for i in range(len(s.q))])

    

    def _copy(s):
        return trajmute()






import matplotlib.pyplot as plt
import numpy as np
import time

class erun:
  
  def __init__(s,vo,obj,population=10,show=True,delay=None):
    """vo: some evo
       obj: some eobj
       population=10: size of population
    
    """
    s.vo=vo
    s.obj=obj
    
    s.vo.populate(s.obj,population)
    
    s.shallmaximize=s.obj.shallmaximize()
    
    s.delay=delay
    
    s.show=show
    
    s.maxs=[]
    s.means=[]
    s.stds=[]
  def getwinner(s):
    return s.vo.getwinner()
  def gettopn(s,n=3):
    return s.vo.gettopn(n=n)
  
  def run(s,maxsteps=1000000,goalstrength=1000000.0):
    for i in range(maxsteps):
      mx,mm,ss=s.vo.logeneration(s.show)
      s.maxs.append(mx)
      s.means.append(mm)
      s.stds.append(ss)
      if s.shallmaximize:
        if mx>=goalstrength:return
      else:
        if mx<=goalstrength:return
      if not s.delay is None:
        time.sleep(s.delay)
  
  def show_history(s,log=False):
    i=np.arange(len(s.maxs))
    
    plt.close()
    
    lab="max"
    if not s.shallmaximize:lab="min"
    plt.plot(i,s.maxs,alpha=0.5,color="red",label=lab)
    
    plt.plot(i,s.means,alpha=0.5,color="black",label="mean")
    me=np.array(s.means)
    st=np.array(s.stds)
    plt.fill_between(i,me-st,me+st,color="gray",alpha=0.3)
    
    if log:plt.yscale("log",nonpositive="clip")
    
    plt.legend()
    
    plt.show()
      











from xevo.eobj import eobj

import random
from abc import abstractmethod

class eobjplus(eobj):
  """eobj, but multiple functions"""
  def initial(s):
    eobj.initial(s)
    s.func=[]#list of dictionaries of type {"f":func()->eobjplus,"ws":probablility}
    s.regall()
  
  @abstractmethod
  def regall(s):
    """register all functions, needs to be overwritten"""
    pass
  
  def register(s,f,ws=1.0):
    s.func.append({"f":f,"ws":ws})
  
  def __init__(s,q=None):
    s.initial()
    s.q=q

  def _mutbyfunc(s,func):
    sm=0.0
    for fun in func:
      sm+=fun["ws"]
    r=random.random()*sm/1.00001
    for fun in func:
      if r<fun["ws"]:
        return fun["f"]()
      else:
        r-=fun["ws"]
    if len(fun)>0:
      return random.choice(fun)["f"]()
    else:
      return random.choice(s.func)["f"]()

  def mutate(s):
    return s._mutbyfunc(s.func)
  
  def copy(s):
    ret=s._copy()
    ret.regall()
    return ret





from xevo.eobjplus import eobjplus

import numpy as np

class pionplus(eobjplus):
  """trivial test eobj trying to find the exact value of pi"""
  def __init__(s,q=3):
    s.initial()
    s.q=q

  def __str__(s):
    return str(s.q)
  def __add__(a,b):
    return pionplus((a.q+b.q)/2)
  def shallmaximize(s):return False

  def regall(s):
    s.func=[]
    s.register(s.mutate1)
    s.register(s.mutate2)
    s.register(s.mutate3)
    s.register(s.mutate4)

  
  def randomize(s):
    return pionplus(np.random.random()*2+2)
  def mutate1(s):
    sn=s.q
    alpha=np.random.random()
    power=-10*np.random.random()
    sn+=alpha*np.exp(power)
    return pionplus(sn)
  def mutate2(s):
    sn=s.q
    alpha=-np.random.random()
    power=-10*np.random.random()
    sn+=alpha*np.exp(power)
    return pionplus(sn)
  def mutate3(s):
    sn=s.q
    alpha=2*np.random.random()-1
    power=-5-5*np.random.random()
    sn+=alpha*np.exp(power)
    return pionplus(sn)
  def mutate4(s):
    sn=s.q
    alpha=2*np.random.random()-1
    power=-5*np.random.random()
    sn+=alpha*np.exp(power)
    return pionplus(sn)
  
  def calcstrength(s):
    return np.abs(s.q-np.pi)
  def _copy(s):
    return pionplus(s.q)



from xevo.objects.bitflip import bitflip
from xevo.objects.dualflip import dualflip
from xevo.objects.gotozero import gotozero
from xevo.objects.oion import oion
from xevo.objects.pionplus import pionplus
from xevo.objects.pion import pion




from xevo.evo import evo

import numpy as np
import time
import random

from xevo.optimizers.trivmute import *


class trivmutetraj(trivmute):
    """trivmute, but for traj applications"""
    def __init__(s,forget=0):
        s.initial()
        s.forget=forget



    def generation(s)->None:
        os=s.q.strength()
        n=s.q.mutate()
        ns=n.strength()
        if (ns<=os) != n.shallmaximize():
            s.q=n
        else:s.forget+=1

    


    def _copy(s):
        return trivmutetraj(forget=s.forget)

    def __str__(s):
        return "trivmutetraj in "+str(s.forget)
    def __repr__(s):
        return str(s)





from xevo.evo import evo

import numpy as np
import time
        

class morthoevo(evo):
    """orthoevo, but using highdimensional definition of similarity. Does not neccesarily produce better results than orthoevo, since orthoevo seems to produce similar groups in higher dimensions (if d1 is 3 than d3=9)"""
    
    def __init__(s,wsmerge=0.3):
        s.initial()
        s.wsmerge=wsmerge



    def generation(s)->None:
        acd=np.random.randint(s.q[0].orthodim())
        s.q.sort(key=lambda x:x.ortho()[acd])
        i1=np.random.randint(len(s.q)-1)
        i2=i1+1
        i1won=s.q[i1].figth(s.q[i2])
        iwon=i1
        ilost=i2
        if not i1won:
            iwon=i2
            ilost=i1
        if np.random.random()<s.wsmerge:
            s.q[ilost]+=s.q[iwon]
        else:
            s.q[ilost]=s.q[iwon].mutate()
        
        


    def _copy(s):
        ret= morthoevo(s.wsmerge)
        return ret





from abc import ABCMeta,abstractmethod

from time import time
import numpy as np

class evo(object):
  """
  any evo class contains at least
  def generation(s)->None:
  def _copy(s) -> "subclass of evo":
  
  """
  __metaclass__=ABCMeta
  
  def initial(s):
    s.q=[]
    s.i=0
  def __init__(s):
    s.initial()
  
  def populate(s,person,n=10)->None:
    s.q.append(person)
    while len(s.q)<n:
      s.q.append(person.randomize())

  
  @abstractmethod
  def generation(s)->None:
    """runs one update step"""
    pass
  
  def winner(s)->int:
    """finds the winner of the current batch"""
    val=[qq.strength() for qq in s.q]
    if s.q[0].shallmaximize():
      return np.argmax(val)
    else:
      return np.argmin(val)
  def topn(s,n=3)->"[int]":
    """finds n top players of the current batch"""
    if s.q[0].shallmaximize():
      val=[qq.strength() for qq in s.q]
    else:
      val=[-qq.strength() for qq in s.q]
    val=np.array(val)

    return val.argsort()[-n:][::-1]
  
  def average(s)->"float,float":
    """finds the average strength and its standard deviation"""
    val=[qq.strength() for qq in s.q]
    return np.mean(val),np.std(val)

  
  def logeneration(s,show=True)->"max,mean,std strength":
    """same as generation, but prints some stuff about the current gen"""
    t0=time()
    s.generation()
    t1=time()
    dt=int(t1-t0)
    index=s.winner()
    win=s[index]
    strength=win.strength()
    s.i+=1
    mm,ss=s.average()

    if show:
      print(f"-----generation {s.i} needed {dt}s-----")
      print(f"best object {index} reached strength {strength}")
      print(f"mean strength {mm} +- {ss}")
      print("")
    
    return strength,mm,ss
    
    
  
  def __getitem__(s, key):
    return s.q[key]
  
  def __setitem__(s,key,obj):
    s.q[key]=obj

  @abstractmethod
  def _copy(s) -> "subclass of evo":
    """copies everything used in the current object"""
    pass
  def copy(s) -> "subclass of evo":
    ret=s._copy()
    ret.i=s.i
    ret.q=[]
    for qq in s.q:
      ret.q.append(qq.copy())
    return ret
  
  def getwinner(s)->"eobj":
    return s[s.winner()]
  def gettopn(s,n=3)->"[eobj]":
    return [s[q] for q in s.topn(n=n)]
  


from xevo.evo import evo

import numpy as np
import time
    

class crossevo(evo):
  """similar to lineevo but no concept of neighbours"""
  
  def __init__(s,wsmerge=0.3):
    s.initial()
    s.wsmerge=wsmerge



  def generation(s)->None:
    i1=np.random.randint(len(s.q))
    i2=i1
    while i1==i2:
      i2=np.random.randint(len(s.q))
    i1won=s.q[i1].figth(s.q[i2])
    iwon=i1
    ilost=i2
    if not i1won:
      iwon=i2
      ilost=i1
    if np.random.random()<s.wsmerge:
      s.q[ilost]+=s.q[iwon]
    else:
      s.q[ilost]=s.q[iwon].mutate()
    
    


  def _copy(s):
    return crossevo(s.wsmerge)






from xevo.eobj import eobj

import numpy as np
import time

class gotozero(eobj):
  """eobj that tries to move nd ints to zero"""
  def __init__(s,q=None):
    if q is None:q=[10,10]
    #print("initialising gotozero with")
    #print(q)
    #time.sleep(1)
    s.initial()
    s.q=q
  def __str__(s):
    return str(s.q)
  def __add__(a,b):
    return gotozero([aa+bb for aa,bb in zip(a,b)])
  def shallmaximize(s):return False

  
  def randomize(s):
    return s.copy()#that kind of a bodge i guess
    return gotozero([np.random.randint(-10,11) for i in range(len(s.q))])
  def mutate(s):
    s0=s
    s=s.copy()
    i=np.random.randint(len(s.q))
    s.q[i]+=np.random.choice([-1,1])
    return s
  def calcstrength(s):
    return np.sum([zw**2 for zw in s.q])
  def _copy(s):
    return gotozero(q=[zw for zw in s.q])
